using System;
using System.Collections.Generic;
using System.Text;

namespace FlutterCodeGenerator
{
    public class ModelType
    {
        public Type ParameterType { get; set; }

        /// <summary>
        ///Name(First Character Lower Case)
        /// </summary>
        public string Name_Lower { get; set; }

        /// <summary>
        ///Name(First Character Upper Case)
        /// </summary>
        public string Name_Upper { get; set; }

        public string DefaultValue { get; set; }

        public ModelType(Type type, string name)
        {
            ParameterType = type;
            Name_Lower = LetterConverterHelper.FirstCharToLower(name);
            Name_Upper = LetterConverterHelper.FirstCharToUpper(name);
            DefaultValue = null;
        }

        public virtual string GetFlutterTypeName(bool isDefault = true, bool isGenericName = false, bool isSingle = false)
        {
            return ParameterType.Name;
        }

        public virtual string GetDartString()
        {
            return string.Empty;
        }

        protected string QuestionMarkCheck(ModelType child)
        {
            if (child.DefaultValue != null)
            {
                return "";
            }
            else
            {
                return "?";
            }
        }
    }

    public class SimpleModelType : ModelType
    {
        public SimpleModelType(Type type, string name) : base(type, name)
        {
        }
    }

    public class StringModeType : SimpleModelType
    {
        public StringModeType(string name) : base(typeof(string), name)
        {
        }
    }

    public class DateTimeModeType : SimpleModelType
    {
        public DateTimeModeType(string name) : base(typeof(DateTime), name)
        {
        }
    }

    public class IntModeType : SimpleModelType
    {
        public IntModeType(string name) : base(typeof(int), name)
        {
            DefaultValue = "0";
        }

        public override string GetFlutterTypeName(bool isDefault = true, bool isGenericName = false, bool isSingle = false)
        {
            return "int";
        }
    }

    public class DoubleModeType : SimpleModelType
    {
        public DoubleModeType(string name) : base(typeof(int), name)
        {
            DefaultValue = "0";
        }

        public override string GetFlutterTypeName(bool isDefault = true, bool isGenericName = false, bool isSingle = false)
        {
            return "double";
        }
    }

    public class BoolModeType : SimpleModelType
    {
        public BoolModeType(string name) : base(typeof(bool), name)
        {
            DefaultValue = "false";
        }

        public override string GetFlutterTypeName(bool isDefault = true, bool isGenericName = false, bool isSingle = false)
        {
            return "bool";
        }
    }

    public class ComplexModelType : ModelType
    {
        public List<ModelType> Children { get; set; }

        public ComplexModelType(Type type, string name) : base(type, name)
        {
            Children = new List<ModelType>();
        }

        protected string ToJson(List<ModelType> Children)
        {
            var source = new StringBuilder();
            source.AppendLine("\tMap<String, dynamic> toJson() {");
            source.AppendLine("\t\tfinal map = Map<String, dynamic>();");
            foreach (var child in Children)
            {
                if (child is EnumModeType)
                {
                    source.AppendLine($"\t\tmap['{child.Name_Upper}'] = {child.Name_Lower}.index;");
                }
                else if (child.DefaultValue != null)
                {
                    source.AppendLine($"\t\tmap['{child.Name_Upper}'] = {child.Name_Lower};");
                }
                else
                {
                    source.AppendLine($"\t\tif({child.Name_Lower} != null)");
                    source.AppendLine($"\t\t\tmap['{child.Name_Upper}'] = {child.Name_Lower};");
                }
            }
            source.AppendLine("\t\treturn map;");
            source.AppendLine("\t}");
            return source.ToString();
        }

        protected string FromJson(List<ModelType> Children)
        {
            var source = new StringBuilder();
            string flutterTypeName = null;
            if (this is UserDefinedGenericComplexModelType)
            {
                flutterTypeName = GetFlutterTypeName(false, false, true);
            }
            else
            {
                flutterTypeName = GetFlutterTypeName();
            }

            source.AppendLine($"\tfactory {flutterTypeName}.fromJson(Map<String, dynamic> map) {{");
            foreach (var child in Children)
            {
                if (child is ArrayModelType)
                {
                    source.AppendLine($"\t\tfinal {child.Name_Lower}Data = map['{child.Name_Upper}'];");
                }
                else if (child is ListModelType listModelType)
                {
                    if (this is UserDefinedGenericComplexModelType type)
                    {
                        if (type.GenericArgumentModelType.ParameterType == listModelType.GenericArgumentModelType.ParameterType)
                        {
                            source.AppendLine($"\t\tList<T> {listModelType.Name_Lower}List = [];");
                            source.AppendLine($"\t\tif (map['{listModelType.Name_Upper}'] != null) {{");
                            source.AppendLine($"\t\t\t{listModelType.Name_Lower}List.addAll(");
                            source.AppendLine($"\t\t\t\t\t(map['{listModelType.Name_Upper}'] as List).map((e) => FJsonConvert.fromJson<T>(e)!));");
                            source.AppendLine($"\t\t}}");
                        }
                    }
                }
                else if (child is UserDefinedGenericComplexModelType userDefinedGenericComplexModelType)
                {
                    if (this is UserDefinedGenericComplexModelType type)
                    {
                        if (type.GenericArgumentModelType.ParameterType == userDefinedGenericComplexModelType.GenericArgumentModelType.ParameterType)
                        {
                            source.AppendLine($"\t\tList<T> {userDefinedGenericComplexModelType.Name_Lower}List = [];");
                            source.AppendLine($"\t\tif (map['{userDefinedGenericComplexModelType.Name_Upper}'] != null) {{");
                            source.AppendLine($"\t\t\t{userDefinedGenericComplexModelType.Name_Lower}List.addAll(");
                            source.AppendLine($"\t\t\t\t\t(map['{userDefinedGenericComplexModelType.Name_Upper}'] as List).map((e) => FJsonConvert.fromJson<T>(e)!));");
                            source.AppendLine($"\t\t}}");
                        }
                    }
                }
            }
            source.AppendLine($"\t\treturn {flutterTypeName}( ");
            foreach (var child in Children)
            {
                if (child is EnumModeType)
                {
                    source.AppendLine($"\t\t\t{child.Name_Lower}: {child.GetFlutterTypeName()}.values.firstWhere((e) => e.index == map['{child.Name_Upper}']),");
                }
                else if (child is ArrayModelType)
                {

                    var arrayTypeName = child.ParameterType.FullName.Replace("[]", string.Empty);
                    var arrayType = child.ParameterType.Assembly.GetType(arrayTypeName);
                    var modelType = ModelTypeGenerator.Create(arrayType, "");
                    if (modelType is SimpleModelType)
                    {
                        source.AppendLine($"\t\t\t{child.Name_Lower}: {child.Name_Lower}Data != null ? ({child.Name_Lower}Data as List).map((e) => e as {child.ParameterType.Name.Replace("[]", string.Empty)}).toList(): const [],");
                    }
                    else
                    {
                        source.AppendLine($"\t\t\t{child.Name_Lower}: {child.Name_Lower}Data != null ? ({child.Name_Lower}Data as List).map((e) => {child.ParameterType.Name.Replace("[]", string.Empty)}.fromJson(e)).toList(): const [],");
                    }
                }
                else if (child is DateTimeModeType)
                {
                    source.AppendLine($"\t\t\t{child.Name_Lower}: map['{child.Name_Upper}'] != null ? DateTime.parse(map['{child.Name_Upper}']) : null,");
                }
                else if (child is ListModelType listModelType)
                {
                    if (this is UserDefinedGenericComplexModelType type)
                    {
                        if (type.GenericArgumentModelType.ParameterType == listModelType.GenericArgumentModelType.ParameterType)
                        {
                            source.AppendLine($"\t\t\t{child.Name_Lower}: {listModelType.Name_Lower}List,");
                        }
                        else if (listModelType.GenericArgumentModelType is SimpleModelType)
                        {
                            var model = listModelType.GenericArgumentModelType.GetFlutterTypeName();
                            source.AppendLine($"\t\t\t{child.Name_Lower}: map['{child.Name_Upper}'].cast<{model}>().toList(),");
                        }
                        else
                        {
                            var model = listModelType.GenericArgumentModelType.GetFlutterTypeName();
                            source.AppendLine($"\t\t\t{child.Name_Lower}: map['{child.Name_Upper}'].map((e)=>{model}.fromJson(e as Map<String,dynamic>)).toList(),");
                        }
                    }
                    else if (this is UserDefinedComplexModeType)
                    {
                        if (listModelType.GenericArgumentModelType is SimpleModelType)
                        {
                            var model = listModelType.GenericArgumentModelType.GetFlutterTypeName();
                            source.AppendLine($"\t\t\t{child.Name_Lower}: map['{child.Name_Upper}'].cast<{model}>().toList(),");
                        }
                        else
                        {
                            var model = listModelType.GenericArgumentModelType.GetFlutterTypeName();
                            source.AppendLine($"\t\t\t{child.Name_Lower}: map['{child.Name_Upper}'].map((e)=>{model}.fromJson(e as Map<String,dynamic>)).toList(),");
                        }
                    }
                }
                else if (child is DictionaryModelType dictionaryModelType)
                {
                    var argumentName1 = dictionaryModelType.GenericArgumentModelTypeList[0].GetFlutterTypeName();
                    var argumentName2 = dictionaryModelType.GenericArgumentModelTypeList[1].GetFlutterTypeName();
                    source.AppendLine($"\t\t\t{child.Name_Lower}: map['{child.Name_Upper}'].cast<{argumentName1},{argumentName2}>(),");
                }
                else if (child is UserDefinedGenericComplexModelType userDefinedGenericComplexModelType)
                {
                    if (this is UserDefinedGenericComplexModelType type)
                    {
                        if (type.GenericArgumentModelType.ParameterType == userDefinedGenericComplexModelType.GenericArgumentModelType.ParameterType)
                        {
                            source.AppendLine($"\t\t\t{child.Name_Lower}: {userDefinedGenericComplexModelType.Name_Lower}List,");
                        }
                        else if (userDefinedGenericComplexModelType.GenericArgumentModelType is SimpleModelType)
                        {
                            var model = userDefinedGenericComplexModelType.GenericArgumentModelType.GetFlutterTypeName();
                            source.AppendLine($"\t\t\t{child.Name_Lower}: map['{child.Name_Upper}'].cast<{model}>().toList(),");
                        }
                        else
                        {
                            var model = userDefinedGenericComplexModelType.GenericArgumentModelType.GetFlutterTypeName();
                            source.AppendLine($"\t\t\t{child.Name_Lower}: map['{child.Name_Upper}'].map((e)=>{model}.fromJson(e as Map<String,dynamic>)).toList(),");
                        }
                    }
                    else if (this is UserDefinedComplexModeType)
                    {
                        if (userDefinedGenericComplexModelType.GenericArgumentModelType is SimpleModelType)
                        {
                            var model = userDefinedGenericComplexModelType.GenericArgumentModelType.GetFlutterTypeName();
                            source.AppendLine($"\t\t\t{child.Name_Lower}: map['{child.Name_Upper}'].cast<{model}>().toList(),");
                        }
                        else
                        {
                            var model = userDefinedGenericComplexModelType.GenericArgumentModelType.GetFlutterTypeName();
                            source.AppendLine($"\t\t\t{child.Name_Lower}: map['{child.Name_Upper}'].map((e)=>{model}.fromJson(e as Map<String,dynamic>)).toList(),");
                        }
                    }
                }
                else
                {
                    source.AppendLine($"\t\t\t{child.Name_Lower}: map['{child.Name_Upper}'],");
                }
            }
            source.AppendLine("\t\t);");
            source.AppendLine("\t}");
            source.AppendLine();
            return source.ToString();
        }
    }

    public class EnumModeType : ComplexModelType
    {
        public EnumModeType(Type type, string name) : base(type, name)
        {
            var firstChildName = Enum.GetNames(type)[0];
            DefaultValue = $"{GetFlutterTypeName()}.{firstChildName}";
        }

        public override string GetDartString()
        {
            var dartString = new StringBuilder();
            dartString.AppendLine($"enum {GetFlutterTypeName()} {{");
            foreach (var enumvalue in Enum.GetNames(ParameterType))
            {
                dartString.AppendLine($"\t{enumvalue},");
            }
            dartString.AppendLine("}");
            return dartString.ToString();
        }
    }

    public class ArrayModelType : SimpleModelType
    {
        public ArrayModelType(Type type, string name) : base(type, name)
        {
            var arrayTypeName = ParameterType.FullName.Replace("[]", string.Empty);
            var arrayType = ParameterType.Assembly.GetType(arrayTypeName);
            ModelTypeGenerator.Create(arrayType, "");
        }

        public override string GetFlutterTypeName(bool isDefault = true, bool isGenericName = false, bool isSingle = false)
        {
            return $"List<{ParameterType.Name[0..^2]}>";
        }
    }

    public class ListModelType : SimpleModelType
    {
        public ModelType GenericArgumentModelType;

        public ListModelType(Type type, string name) : base(type, name)
        {
            var modelType = ModelTypeGenerator.Create(type.GetGenericArguments()[0], "");
            GenericArgumentModelType = modelType;
        }

        public override string GetFlutterTypeName(bool isDefault = true, bool isGenericName = false, bool isSingle = false)
        {
            if (isDefault)
            {
                return ParameterType.Name[0..^2] + "<" + GenericArgumentModelType.GetFlutterTypeName() + ">";
            }

            if (isGenericName)
            {
                return ParameterType.Name[0..^2] + "<T>";
            }
            if (isSingle)
            {
                return ParameterType.Name[0..^2];
            }

            return ParameterType.Name[0..^2] + "<" + GenericArgumentModelType.GetFlutterTypeName() + ">";
        }
    }

    /// <summary>
    /// Dictionary Type
    /// </summary>
    public class DictionaryModelType : SimpleModelType
    {
        public List<ModelType> GenericArgumentModelTypeList;

        public DictionaryModelType(Type type, string name) : base(type, name)
        {
            GenericArgumentModelTypeList = new List<ModelType>();
            foreach (var argumentType in type.GetGenericArguments())
            {
                var modelType = ModelTypeGenerator.Create(argumentType, "");
                GenericArgumentModelTypeList.Add(modelType);
            }
        }

        public override string GetFlutterTypeName(bool isDefault = true, bool isGenericName = false, bool isSingle = false)
        {
            var flutterTypeName = ParameterType.Name[0..^2].Replace("Dictionary", "Map");
            if (isDefault)
            {
                flutterTypeName += "<" + GenericArgumentModelTypeList[0].GetFlutterTypeName() + "," + GenericArgumentModelTypeList[1].GetFlutterTypeName() + ">";
                return flutterTypeName;
            }
            if (isSingle)
            {
                return ParameterType.Name[0..^2].Replace("Dictionary", "Map");
            }
            flutterTypeName += "<" + GenericArgumentModelTypeList[0].GetFlutterTypeName() + "," + GenericArgumentModelTypeList[1].GetFlutterTypeName() + ">";
            return flutterTypeName;
        }
    }

    /// <summary>
    /// User Defined Generic Complex Model Type
    /// </summary>
    public class UserDefinedGenericComplexModelType : ComplexModelType
    {
        public ModelType GenericArgumentModelType;

        public UserDefinedGenericComplexModelType(Type type, string name) : base(type, name)
        {
            var properties = type.GetProperties();
            foreach (var property in properties)
            {
                var child = ModelTypeGenerator.Create(property.PropertyType, property.Name);
                Children.Add(child);
            }
            var modelType = ModelTypeGenerator.Create(type.GetGenericArguments()[0], "");
            GenericArgumentModelType = modelType;
        }

        public override string GetFlutterTypeName(bool isDefault = true, bool isGenericName = false, bool isSingle = false)
        {
            if (isDefault)
            {
                return ParameterType.Name[0..^2] + "<" + GenericArgumentModelType.GetFlutterTypeName() + ">";
            }

            if (isGenericName)
            {
                return ParameterType.Name[0..^2] + "<T>";
            }
            if (isSingle)
            {
                return ParameterType.Name[0..^2];
            }

            return ParameterType.Name[0..^2] + "<" + GenericArgumentModelType.GetFlutterTypeName() + ">";
        }

        public override string GetDartString()
        {
            var dartString = new StringBuilder();
            dartString.AppendLine($"class {GetFlutterTypeName(false, true)} {{");
            foreach (var child in Children)
            {
                string argumentName;
                var questionMark = QuestionMarkCheck(child);

                if (child is ListModelType listModelType)
                {
                    argumentName = GenericArgumentModelType.ParameterType == listModelType.GenericArgumentModelType.ParameterType ? listModelType.GetFlutterTypeName(false, true) : listModelType.GetFlutterTypeName();
                }
                else if (child is UserDefinedGenericComplexModelType userDefinedGenericComplexModelType)
                {
                    argumentName = GenericArgumentModelType.ParameterType == userDefinedGenericComplexModelType.GenericArgumentModelType.ParameterType ? userDefinedGenericComplexModelType.GetFlutterTypeName(false, true) : userDefinedGenericComplexModelType.GetFlutterTypeName();
                }
                else
                {
                    argumentName = child.GetFlutterTypeName();
                }
                dartString.AppendLine($"\t{argumentName}{questionMark} {child.Name_Lower};");
            }
            dartString.AppendLine();
            dartString.AppendLine($"\t{GetFlutterTypeName(false, false, true)}({{");
            foreach (var child in Children)
            {
                if (child.DefaultValue != null)
                {
                    dartString.AppendLine($"\t\tthis.{child.Name_Lower}={child.DefaultValue},");
                }
                else
                {
                    dartString.AppendLine($"\t\tthis.{child.Name_Lower},");
                }
            }
            dartString.AppendLine("\t});");
            dartString.AppendLine();
            dartString.Append(FromJson(Children));
            dartString.Append(ToJson(Children));
            dartString.AppendLine("}");
            return dartString.ToString();
        }
    }

    public class UserDefinedComplexModeType : ComplexModelType
    {
        public UserDefinedComplexModeType(Type type, string name) : base(type, name)
        {
            var properties = type.GetProperties();
            foreach (var property in properties)
            {
                var child = ModelTypeGenerator.Create(property.PropertyType, property.Name);
                Children.Add(child);
            }
        }

        public override string GetFlutterTypeName(bool isDefault = true, bool isGenericName = false, bool isSingle = false)
        {
            return $"{ParameterType.Name}";
        }

        public override string GetDartString()
        {
            var dartString = new StringBuilder();
            dartString.AppendLine($"class {GetFlutterTypeName()} {{");
            foreach (var child in Children)
            {
                var questionMark = QuestionMarkCheck(child);
                if (child is ListModelType || child is DictionaryModelType || child is UserDefinedGenericComplexModelType)
                {
                    dartString.AppendLine($"\t{child.GetFlutterTypeName(false, false)}{questionMark} {child.Name_Lower};");
                }
                else
                {
                    dartString.AppendLine($"\t{child.GetFlutterTypeName()}{questionMark} {child.Name_Lower};");
                }
            }
            dartString.AppendLine();
            dartString.AppendLine($"\t{GetFlutterTypeName()}({{");
            foreach (var child in Children)
            {
                if (child.DefaultValue != null)
                {
                    dartString.AppendLine($"\t\tthis.{child.Name_Lower}={child.DefaultValue},");
                }
                else
                {
                    dartString.AppendLine($"\t\tthis.{child.Name_Lower},");
                }
            }
            dartString.AppendLine("\t});");
            dartString.AppendLine();

            dartString.Append(FromJson(Children));
            dartString.Append(ToJson(Children));
            dartString.AppendLine("}");
            return dartString.ToString();
        }
    }
}
